"""Dark magics about variable names in python"""

from .utils import (
    config,
    VarnameException,
    VarnameRetrievingError,
    ImproperUseError,
    QualnameNonUniqueError,
    VarnameWarning,
    MultiTargetAssignmentWarning,
    MaybeDecoratedFunctionWarning,
    UsingExecWarning,
)
from .core import varname, nameof, will, argname

__version__ = "0.15.1"



"""Some helper functions builtin based upon core features"""
from __future__ import annotations

import inspect
from functools import partial, wraps
from os import PathLike
from typing import Any, Callable, Dict, Tuple, Type, Union

from .utils import IgnoreType
from .ignore import IgnoreList
from .core import argname, varname


def register(
    cls_or_func: type = None,
    frame: int = 1,
    ignore: IgnoreType = None,
    multi_vars: bool = False,
    raise_exc: bool = True,
    strict: bool = True,
) -> Union[Type, Callable]:
    """A decorator to register __varname__ to a class or function

    When registered to a class, it can be accessed by `self.__varname__`;
    while to a function, it is registered to globals, meaning that it can be
    accessed directly.

    Args:
        frame: The call stack index, indicating where this class
            is instantiated relative to where the variable is finally retrieved
        multi_vars: Whether allow multiple variables on left-hand side (LHS).
            If `True`, this function returns a tuple of the variable names,
            even there is only one variable on LHS.
            If `False`, and multiple variables on LHS, a
            `VarnameRetrievingError` will be raised.
        raise_exc: Whether we should raise an exception if failed
            to retrieve the name.
        strict: Whether to only return the variable name if the result of
            the call is assigned to it directly.

    Examples:
        >>> @varname.register
        >>> class Foo: pass
        >>> foo = Foo()
        >>> # foo.__varname__ == 'foo'
        >>>
        >>> @varname.register
        >>> def func():
        >>>   return __varname__
        >>> foo = func() # foo == 'foo'

    Returns:
        The wrapper function or the class/function itself
        if it is specified explictly.
    """
    if inspect.isclass(cls_or_func):
        orig_init = cls_or_func.__init__  # type: ignore

        @wraps(cls_or_func.__init__)  # type: ignore
        def wrapped_init(self, *args, **kwargs):
            """Wrapped init function to replace the original one"""
            self.__varname__ = varname(
                frame - 1,
                ignore=ignore,
                multi_vars=multi_vars,
                raise_exc=raise_exc,
                strict=strict,
            )
            orig_init(self, *args, **kwargs)

        cls_or_func.__init__ = wrapped_init  # type: ignore
        return cls_or_func

    if inspect.isfunction(cls_or_func):

        @wraps(cls_or_func)
        def wrapper(*args, **kwargs):
            """The wrapper to register `__varname__` to a function"""
            cls_or_func.__globals__["__varname__"] = varname(
                frame - 1,
                ignore=ignore,
                multi_vars=multi_vars,
                raise_exc=raise_exc,
                strict=strict,
            )

            try:
                return cls_or_func(*args, **kwargs)
            finally:
                del cls_or_func.__globals__["__varname__"]

        return wrapper

    # None, meaning we have other arguments
    return partial(
        register,
        frame=frame,
        ignore=ignore,
        multi_vars=multi_vars,
        raise_exc=raise_exc,
        strict=strict,
    )


class Wrapper:
    """A wrapper with ability to retrieve the variable name

    Examples:
        >>> foo = Wrapper(True)
        >>> # foo.name == 'foo'
        >>> # foo.value == True

        >>> val = {}
        >>> bar = Wrapper(val)
        >>> # bar.name == 'bar'
        >>> # bar.value is val

    Args:
        value: The value to be wrapped
        raise_exc: Whether to raise exception when varname is failed to retrieve
        strict: Whether to only return the variable name if the wrapper is
            assigned to it directly.

    Attributes:
        name: The variable name to which the instance is assigned
        value: The value this wrapper wraps
    """

    def __init__(
        self,
        value: Any,
        frame: int = 1,
        ignore: IgnoreType = None,
        raise_exc: bool = True,
        strict: bool = True,
    ):
        # This call is ignored, since it's inside varname
        self.name = varname(
            frame=frame - 1,
            ignore=ignore,
            raise_exc=raise_exc,
            strict=strict,
        )
        self.value = value

    def __str__(self) -> str:
        return repr(self.value)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"(name={self.name!r}, value={self.value!r})>"
        )


def jsobj(
    *args: Any,
    vars_only: bool = True,
    frame: int = 1,
    **kwargs: Any,
) -> Dict[str, Any]:
    """A wrapper to create a JavaScript-like object

    When an argument is passed as positional argument, the name of the variable
    will be used as the key, while the value will be used as the value.

    Examples:
        >>> obj = jsobj(a=1, b=2)
        >>> # obj == {'a': 1, 'b': 2}
        >>> # obj.a == 1
        >>> # obj.b == 2
        >>> a = 1
        >>> b = 2
        >>> obj = jsobj(a, b, c=3)
        >>> # obj == {'a': 1, 'b': 2, 'c': 3}

    Args:
        *args: The positional arguments
        vars_only: Whether to only include variables in the output
        frame: The call stack index. You can understand this as the number of
            wrappers around this function - 1.
        **kwargs: The keyword arguments

    Returns:
        A dict-like object
    """
    argnames: Tuple[str, ...] = argname(
        "args",
        vars_only=vars_only,
        frame=frame,
    )  # type: ignore
    out = dict(zip(argnames, args))
    out.update(kwargs)
    return out


def debug(
    var,
    *more_vars,
    prefix: str = "DEBUG: ",
    merge: bool = False,
    repr: bool = True,
    sep: str = "=",
    vars_only: bool = False,
) -> None:
    """Print variable names and values.

    Examples:
        >>> a = 1
        >>> b = object
        >>> print(f'a={a}') # previously, we have to do
        >>> print(f'{a=}')  # or with python3.8
        >>> # instead we can do:
        >>> debug(a) # DEBUG: a=1
        >>> debug(a, prefix='') # a=1
        >>> debug(a, b, merge=True) # a=1, b=<object object at 0x2b9a4c89cf00>

    Args:
        var: The variable to print
        *more_vars: Other variables to print
        prefix: A prefix to print for each line
        merge: Whether merge all variables in one line or not
        sep: The separator between the variable name and value
        repr: Print the value as `repr(var)`? otherwise `str(var)`
    """
    var_names = argname("var", "*more_vars", vars_only=vars_only, func=debug)

    values = (var, *more_vars)
    name_and_values = [
        f"{var_name}{sep}{value!r}" if repr else f"{var_name}{sep}{value}"
        for var_name, value in zip(var_names, values)  # type: ignore
    ]

    if merge:
        print(f"{prefix}{', '.join(name_and_values)}")
    else:
        for name_and_value in name_and_values:
            print(f"{prefix}{name_and_value}")


def exec_code(
    code: str,
    globals: Dict[str, Any] = None,
    locals: Dict[str, Any] = None,
    /,
    sourcefile: PathLike | str = None,
    frame: int = 1,
    ignore: IgnoreType = None,
    **kwargs: Any,
) -> None:
    """Execute code where source code is visible at runtime.

    This function is useful when you want to execute some code, where you want to
    retrieve the AST node of the code at runtime. This function will create a
    temporary file and write the code into it, then execute the code in the
    file.

    Examples:
        >>> from varname import varname
        >>> def func(): return varname()
        >>> exec('var = func()')  # VarnameRetrievingError:
        >>>                       #  Unable to retrieve the ast node.
        >>> from varname.helpers import code_exec
        >>> code_exec('var = func()')  # var == 'var'

    Args:
        code: The code to execute.
        globals: The globals to use.
        locals: The locals to use.
        sourcefile: The source file to write the code into.
            if not given, a temporary file will be used.
            This file will be deleted after the code is executed.
        frame: The call stack index. You can understand this as the number of
            wrappers around this function. This is used to fetch `globals` and
            `locals` from where the destination function (include the wrappers
            of this function)
            is called.
        ignore: The intermediate calls to be ignored. See `varname.ignore`
            Note that if both `globals` and `locals` are given, `frame` and
            `ignore` will be ignored.
        **kwargs: The keyword arguments to pass to `exec`.
    """
    if sourcefile is None:
        import tempfile

        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".py", delete=False
        ) as f:
            f.write(code)
            sourcefile = f.name
    else:
        sourcefile = str(sourcefile)
        with open(sourcefile, "w") as f:
            f.write(code)

    if globals is None or locals is None:
        ignore_list = IgnoreList.create(ignore)
        frame_info = ignore_list.get_frame(frame)
        if globals is None:
            globals = frame_info.f_globals
        if locals is None:
            locals = frame_info.f_locals

    try:
        exec(compile(code, sourcefile, "exec"), globals, locals, **kwargs)
    finally:
        import os

        os.remove(sourcefile)



"""Some internal utilities for varname


Attributes:

    IgnoreElemType: The type for ignore elements
    IgnoreType: The type for the ignore argument
    MODULE_IGNORE_ID_NAME: The name of the ignore id injected to the module.
        Espectially for modules that can't be retrieved by
        `inspect.getmodule(frame)`
"""
import sys
import dis
import ast
import warnings
import inspect
from os import path
from pathlib import Path
from functools import lru_cache, singledispatch
from types import ModuleType, FunctionType, CodeType, FrameType
from typing import Tuple, Union, List, Mapping, Callable, Dict

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias  # pragma: no cover
else:
    from typing import TypeAlias

from executing import Source

OP2MAGIC = {
    ast.Add: "__add__",
    ast.Sub: "__sub__",
    ast.Mult: "__mul__",
    ast.Div: "__truediv__",
    ast.FloorDiv: "__floordiv__",
    ast.Mod: "__mod__",
    ast.Pow: "__pow__",
    ast.LShift: "__lshift__",
    ast.RShift: "__rshift__",
    ast.BitOr: "__or__",
    ast.BitXor: "__xor__",
    ast.BitAnd: "__and__",
    ast.MatMult: "__matmul__",
}

OP2SYMBOL = {
    # BinOp
    ast.Add: "+",
    ast.Sub: "-",
    ast.Mult: "*",
    ast.Div: "/",
    ast.FloorDiv: "//",
    ast.Mod: "%",
    ast.Pow: "**",
    ast.LShift: "<<",
    ast.RShift: ">>",
    ast.BitOr: "|",
    ast.BitXor: "^",
    ast.BitAnd: "&",
    ast.MatMult: "@",
    # BoolOp
    ast.And: "and",
    ast.Or: "or",
    # UnaryOp
    ast.UAdd: "+",
    ast.USub: "-",
    ast.Invert: "~",
    ast.Not: "not ",
    # Compare
    ast.Eq: "==",
    ast.NotEq: "!=",
    ast.Lt: "<",
    ast.LtE: "<=",
    ast.Gt: ">",
    ast.GtE: ">=",
    ast.Is: "is",
    ast.IsNot: "is not",
    ast.In: "in",
    ast.NotIn: "not in",
}

CMP2MAGIC = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__lt__",
    ast.LtE: "__le__",
    ast.Gt: "__gt__",
    ast.GtE: "__ge__",
}

IgnoreElemType = Union[
    # module
    ModuleType,
    # filename of a module
    str,
    Path,
    FunctionType,
    # the module (filename) and qualname
    # If module is None, then all qualname matches the 2nd element
    # will be ignored. Used to ignore <lambda> internally
    Tuple[Union[ModuleType, str], str],
    # Function and number of its decorators
    Tuple[FunctionType, int],
]
IgnoreType = Union[IgnoreElemType, List[IgnoreElemType]]

ArgSourceType: TypeAlias = Union[ast.AST, str]
ArgSourceType: TypeAlias = Union[ArgSourceType, Tuple[ArgSourceType, ...]]
ArgSourceType: TypeAlias = Union[ArgSourceType, Mapping[str, ArgSourceType]]

if sys.version_info >= (3, 8):
    ASSIGN_TYPES = (ast.Assign, ast.AnnAssign, ast.NamedExpr)
    AssignType: TypeAlias = Union[ASSIGN_TYPES]  # type: ignore
else:  # pragma: no cover  # Python < 3.8
    ASSIGN_TYPES = (ast.Assign, ast.AnnAssign)
    AssignType: TypeAlias = Union[ASSIGN_TYPES]  # type: ignore

PY311 = sys.version_info >= (3, 11)
MODULE_IGNORE_ID_NAME = "__varname_ignore_id__"


class config:
    """Global configurations for varname

    Attributes:
        debug: Show debug information for frames being ignored
    """

    debug = False


class VarnameException(Exception):
    """Root exception for all varname exceptions"""


class VarnameRetrievingError(VarnameException):
    """When failed to retrieve the varname"""


class QualnameNonUniqueError(VarnameException):
    """When a qualified name is used as an ignore element but references to
    multiple objects in a module"""


class ImproperUseError(VarnameException):
    """When varname() is improperly used"""


class VarnameWarning(Warning):
    """Root warning for all varname warnings"""


class MaybeDecoratedFunctionWarning(VarnameWarning):
    """When a suspecious decorated function used as ignore function directly"""


class MultiTargetAssignmentWarning(VarnameWarning):
    """When varname tries to retrieve variable name in
    a multi-target assignment"""


class UsingExecWarning(VarnameWarning):
    """When exec is used to retrieve function name for `argname()`"""


@lru_cache()
def cached_getmodule(codeobj: CodeType):
    """Cached version of inspect.getmodule"""
    return inspect.getmodule(codeobj)


def get_node(
    frame: int,
    ignore: IgnoreType = None,
    raise_exc: bool = True,
    ignore_lambda: bool = True,
) -> ast.AST:
    """Try to get node from the executing object.

    This can fail when a frame is failed to retrieve.
    One case should be when python code is executed in
    R pacakge `reticulate`, where only first frame is kept.

    When the node can not be retrieved, try to return the first statement.
    """
    from .ignore import IgnoreList

    ignore = IgnoreList.create(ignore, ignore_lambda=ignore_lambda)
    try:
        frameobj = ignore.get_frame(frame)
    except VarnameRetrievingError:
        return None

    return get_node_by_frame(frameobj, raise_exc)


def get_node_by_frame(frame: FrameType, raise_exc: bool = True) -> ast.AST:
    """Get the node by frame, raise errors if possible"""
    exect = Source.executing(frame)

    if exect.node:
        # attach the frame for better exception message
        # (ie. where ImproperUseError happens)
        exect.node.__frame__ = frame
        return exect.node

    if exect.source.text and exect.source.tree and raise_exc:  # pragma: no cover
        raise VarnameRetrievingError(
            "Couldn't retrieve the call node. "
            "This may happen if you're using some other AST magic at the "
            "same time, such as pytest, ipython, macropy, or birdseye."
        )

    return None


def lookfor_parent_assign(node: ast.AST, strict: bool = True) -> AssignType:
    """Look for an ast.Assign node in the parents"""
    while hasattr(node, "parent"):
        node = node.parent

        if isinstance(node, ASSIGN_TYPES):
            return node

        if strict:
            break
    return None


def node_name(
    node: ast.AST,
    subscript_slice: bool = False,
) -> Union[str, Tuple[Union[str, Tuple], ...]]:
    """Get the node node name.

    Raises ImproperUseError when failed
    """
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Attribute):
        return f"{node_name(node.value)}.{node.attr}"
    if isinstance(node, ast.Constant):
        return repr(node.value)
    if isinstance(node, (ast.List, ast.Tuple)) and not subscript_slice:
        return tuple(node_name(elem) for elem in node.elts)
    if isinstance(node, ast.List):
        return f"[{', '.join(node_name(elem) for elem in node.elts)}]"  # type: ignore
    if isinstance(node, ast.Tuple):
        if len(node.elts) == 1:
            return f"({node_name(node.elts[0])},)"
        return f"({', '.join(node_name(elem) for elem in node.elts)})"  # type: ignore
    if isinstance(node, ast.Starred):
        return f"*{node_name(node.value)}"
    if isinstance(node, ast.Slice):
        return (
            f"{node_name(node.lower)}:{node_name(node.upper)}:{node_name(node.step)}"
            if node.lower is not None
            and node.upper is not None
            and node.step is not None
            else f"{node_name(node.lower)}:{node_name(node.upper)}"
            if node.lower is not None and node.upper is not None
            else f"{node_name(node.lower)}:"
            if node.lower is not None
            else f":{node_name(node.upper)}"
            if node.upper is not None
            else ":"
        )
    if isinstance(node, ast.BinOp):
        return (
            f"{node_name(node.left)} {OP2SYMBOL[type(node.op)]} {node_name(node.right)}"
        )
    if isinstance(node, ast.BoolOp):
        return f" {OP2SYMBOL[type(node.op)]} ".join(
            str(node_name(value)) for value in node.values
        )
    if isinstance(node, ast.Compare):
        parts = [str(node_name(node.left))]
        for op, comparator in zip(node.ops, node.comparators):
            parts.append(OP2SYMBOL[type(op)])
            parts.append(str(node_name(comparator)))
        return " ".join(parts)
    if isinstance(node, ast.UnaryOp):
        return f"{OP2SYMBOL[type(node.op)]}{node_name(node.operand)}"

    name = type(node).__name__
    if isinstance(node, ast.Subscript):
        try:
            return f"{node_name(node.value)}[{node_name(node.slice, True)}]"
        except ImproperUseError:
            name = f"{node_name(node.value)}[{type(node.slice).__name__}]"

    raise ImproperUseError(
        f"Node {name!r} detected, but only following nodes are supported: \n"
        "  - ast.Name (e.g. x)\n"
        "  - ast.Attribute (e.g. x.y, x be other supported nodes)\n"
        "  - ast.Constant (e.g. 1, 'a')\n"
        "  - ast.List (e.g. [x, y, z])\n"
        "  - ast.Tuple (e.g. (x, y, z))\n"
        "  - ast.Starred (e.g. *x)\n"
        "  - ast.Subscript with slice of the above nodes (e.g. x[y])\n"
        "  - ast.Subscript with ast.BinOp/ast.BoolOp/ast.Compare/ast.UnaryOp "
        "(e.g. x[y + 1], x[y and z])"
    )


@lru_cache()
def bytecode_nameof(code: CodeType, offset: int) -> str:
    """Cached Bytecode version of nameof

    We are trying this version only when the sourcecode is unavisible. In most
    cases, this will happen when user is trying to run a script in REPL/
    python shell, with `eval`, or other circumstances where the code is
    manipulated to run but sourcecode is not available.
    """
    kwargs: Dict[str, bool] = (
        {"show_caches": True} if sys.version_info[:2] >= (3, 11) else {}
    )

    instructions = list(dis.get_instructions(code, **kwargs))
    ((current_instruction_index, current_instruction),) = (
        (index, instruction)
        for index, instruction in enumerate(instructions)
        if instruction.offset == offset
    )

    while current_instruction.opname == "CACHE":  # pragma: no cover
        current_instruction_index -= 1
        current_instruction = instructions[current_instruction_index]

    pos_only_error = VarnameRetrievingError(
        "'nameof' can only be called with a single positional argument "
        "when source code is not avaiable."
    )
    if current_instruction.opname in (  # pragma: no cover
        "CALL_FUNCTION_EX",
        "CALL_FUNCTION_KW",
    ):
        raise pos_only_error

    if current_instruction.opname not in (
        "CALL_FUNCTION",
        "CALL_METHOD",
        "CALL",
        "CALL_KW",
    ):
        raise VarnameRetrievingError("Did you call 'nameof' in a weird way?")

    current_instruction_index -= 1
    name_instruction = instructions[current_instruction_index]
    while name_instruction.opname in ("CACHE", "PRECALL"):  # pragma: no cover
        current_instruction_index -= 1
        name_instruction = instructions[current_instruction_index]

    if name_instruction.opname in ("KW_NAMES", "LOAD_CONST"):  # LOAD_CONST python 3.13
        raise pos_only_error

    if not name_instruction.opname.startswith("LOAD_"):
        raise VarnameRetrievingError("Argument must be a variable or attribute")

    name = name_instruction.argrepr
    if not name.isidentifier():
        raise VarnameRetrievingError(
            f"Found the variable name {name!r} which is obviously wrong. "
            "This may happen if you're using some other AST magic at the "
            "same time, such as pytest, ipython, macropy, or birdseye."
        )

    return name


def attach_ignore_id_to_module(module: ModuleType) -> None:
    """Attach the ignore id to module

    This is useful when a module cannot be retrieved by frames using
    `inspect.getmodule`, then we can use this id, which will exist in
    `frame.f_globals` to check if the module matches in ignore.

    Do it only when the __file__ is not avaiable or does not exist for
    the module. Since this probably means the source is not avaiable and
    `inspect.getmodule` would not work
    """
    module_file = getattr(module, "__file__", None)
    if module_file is not None and path.isfile(module_file):
        return
    # or it's already been set
    if hasattr(module, MODULE_IGNORE_ID_NAME):
        return

    setattr(module, MODULE_IGNORE_ID_NAME, f"<varname-ignore-{id(module)})")


def frame_matches_module_by_ignore_id(
    frame: FrameType, module: ModuleType
) -> bool:
    """Check if the frame is from the module by ignore id"""
    ignore_id_attached = getattr(module, MODULE_IGNORE_ID_NAME, object())
    ignore_id_from_frame = frame.f_globals.get(MODULE_IGNORE_ID_NAME, object())
    return ignore_id_attached == ignore_id_from_frame


@lru_cache()
def check_qualname_by_source(
    source: Source, modname: str, qualname: str
) -> None:
    """Check if a qualname in module is unique"""
    if not source.tree:
        # no way to check it, skip
        return
    nobj = list(source._qualnames.values()).count(qualname)
    if nobj > 1:
        raise QualnameNonUniqueError(
            f"Qualname {qualname!r} in "
            f"{modname!r} refers to multiple objects."
        )


def debug_ignore_frame(msg: str, frameinfo: inspect.FrameInfo = None) -> None:
    """Print the debug message for a given frame info object

    Args:
        msg: The debugging message
        frameinfo: The FrameInfo object for the frame
    """
    if not config.debug:
        return
    if frameinfo is not None:
        msg = (
            f"{msg} [In {frameinfo.function!r} at "
            f"{frameinfo.filename}:{frameinfo.lineno}]"
        )
    sys.stderr.write(f"[{__package__}] DEBUG: {msg}\n")


def argnode_source(
    source: Source, node: ast.AST, vars_only: bool
) -> Union[str, ast.AST]:
    """Get the source of an argument node

    Args:
        source: The executing source object
        node: The node to get the source from
        vars_only: Whether only allow variables and attributes

    Returns:
        The source of the node (node.id for ast.Name,
            node.attr for ast.Attribute). Or the node itself if the source
            cannot be fetched.
    """
    if isinstance(node, ast.Constant):
        return repr(node.value)

    if sys.version_info < (3, 9):  # pragma: no cover
        if isinstance(node, ast.Index):
            node = node.value
        if isinstance(node, ast.Num):
            return repr(node.n)
        if isinstance(node, (ast.Bytes, ast.Str)):
            return repr(node.s)
        if isinstance(node, ast.NameConstant):
            return repr(node.value)

    if vars_only:
        return (
            node.id
            if isinstance(node, ast.Name)
            else node.attr
            if isinstance(node, ast.Attribute)
            else node
        )

    # requires asttokens
    return source.asttokens().get_text(node)


def get_argument_sources(
    source: Source,
    node: ast.Call,
    func: Callable,
    vars_only: bool,
) -> Mapping[str, ArgSourceType]:
    """Get the sources for argument from an ast.Call node

    >>> def func(a, b, c, d=4):
    >>>  ...
    >>> x = y = z = 1
    >>> func(y, x, c=z)
    >>> # argument_sources = {'a': 'y', 'b', 'x', 'c': 'z'}
    >>> func(y, x, c=1)
    >>> # argument_sources = {'a': 'y', 'b', 'x', 'c': ast.Num(n=1)}
    """
    # <Signature (a, b, c, d=4)>
    signature = inspect.signature(func, follow_wrapped=False)
    # func(y, x, c=z)
    # ['y', 'x'], {'c': 'z'}
    arg_sources = [
        argnode_source(source, argnode, vars_only) for argnode in node.args
    ]
    kwarg_sources = {
        argnode.arg: argnode_source(source, argnode.value, vars_only)
        for argnode in node.keywords
        if argnode.arg is not None
    }
    bound_args = signature.bind_partial(*arg_sources, **kwarg_sources)
    argument_sources = bound_args.arguments
    # see if *args and **kwargs have anything assigned
    # if not, assign () and {} to them
    for parameter in signature.parameters.values():
        if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            argument_sources.setdefault(parameter.name, ())
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            argument_sources.setdefault(parameter.name, {})
    return argument_sources


def get_function_called_argname(frame: FrameType, node: ast.Call) -> Callable:
    """Get the function who called argname"""
    # variable
    if isinstance(node.func, ast.Name):
        func = frame.f_locals.get(
            node.func.id, frame.f_globals.get(node.func.id)
        )
        if func is None:  # pragma: no cover
            # not sure how it would happen but in case
            raise VarnameRetrievingError(
                f"Cannot retrieve the function by {node.func.id!r}."
            )
        return func

    # use pure_eval
    pure_eval_fail_msg = None
    try:
        from pure_eval import Evaluator, CannotEval
    except ImportError:
        pure_eval_fail_msg = "'pure_eval' is not installed."
    else:
        try:
            evaluator = Evaluator.from_frame(frame)
            return evaluator[node.func]
        except CannotEval:
            pure_eval_fail_msg = (
                f"Cannot evaluate node {ast.dump(node.func)} "
                "using 'pure_eval'."
            )

    # try eval
    warnings.warn(
        f"{pure_eval_fail_msg} "
        "Using 'eval' to get the function that calls 'argname'. "
        "Try calling it using a variable reference to the function, or "
        "passing the function to 'argname' explicitly.",
        UsingExecWarning,
    )
    expr = ast.Expression(node.func)
    code = compile(expr, "<ast-call>", "eval")
    return eval(code, frame.f_globals, frame.f_locals)


@singledispatch
def reconstruct_func_node(node: ast.AST) -> ast.Call:
    """Reconstruct the ast.Call node from

    `x.a` to `x.__getattr__('a')`
    `x.a = b` to `x.__setattr__('a', b)`
    `x[a]` to `x.__getitem__(a)`
    `x[a] = b` to `x.__setitem__(a, 1)`
    `x + a` to `x.__add__(a)`
    `x < a` to `x.__lt__(a)`
    """
    raise VarnameRetrievingError(
        f"Cannot reconstruct ast.Call node from {type(node).__name__}, "
        "expecting Call, Attribute, Subscript, BinOp, Compare."
    )


@reconstruct_func_node.register(ast.Call)
def _(node: ast.Call) -> ast.Call:
    return node


@reconstruct_func_node.register(ast.Attribute)
@reconstruct_func_node.register(ast.Subscript)
def _(node: Union[ast.Attribute, ast.Subscript]) -> ast.Call:
    """Reconstruct the function node for
    `x.__getitem__/__setitem__/__getattr__/__setattr__`"""
    nodemeta = {
        "lineno": node.lineno,
        "col_offset": node.col_offset,
    }
    keynode = (
        node.slice
        if isinstance(node, ast.Subscript)
        else ast.Constant(value=node.attr)
    )

    # x[1], x.a
    if isinstance(node.ctx, ast.Load):
        if PY311:
            return ast.Call(
                func=ast.Attribute(
                    value=node.value,
                    attr=(
                        "__getitem__"
                        if isinstance(node, ast.Subscript)
                        else "__getattr__"
                    ),
                    ctx=ast.Load(),
                    **nodemeta,
                ),
                args=[keynode],
                keywords=[],
            )
        else:  # pragma: no cover
            return ast.Call(  # type: ignore
                func=ast.Attribute(
                    value=node.value,
                    attr=(
                        "__getitem__"
                        if isinstance(node, ast.Subscript)
                        else "__getattr__"
                    ),
                    ctx=ast.Load(),
                    **nodemeta,
                ),
                args=[keynode],
                keywords=[],
                starargs=None,
                kwargs=None,
            )

    # x[a] = b, x.a = b
    if (
        not hasattr(node, "parent")
        or not isinstance(node.parent, ast.Assign)  # type: ignore
        or len(node.parent.targets) != 1  # type: ignore
    ):
        raise ImproperUseError(
            rich_exc_message(
                "Expect `x[a] = b` or `x.a = b` directly, got "
                f"{ast.dump(node)}.",
                node,
            )
        )

    if PY311:
        return ast.Call(
            func=ast.Attribute(
                value=node.value,
                attr=(
                    "__setitem__"
                    if isinstance(node, ast.Subscript)
                    else "__setattr__"
                ),
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[keynode, node.parent.value],  # type: ignore
            keywords=[],
        )
    else:  # pragma: no cover
        return ast.Call(
            func=ast.Attribute(
                value=node.value,
                attr=(
                    "__setitem__"
                    if isinstance(node, ast.Subscript)
                    else "__setattr__"
                ),
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[keynode, node.parent.value],  # type: ignore
            keywords=[],
            starargs=None,
            kwargs=None,
        )


@reconstruct_func_node.register(ast.Compare)
def _(node: ast.Compare) -> ast.Call:
    """Reconstruct the function node for `x < a`"""
    # When the node is identified by executing, len(ops) is always 1.
    # Otherwise, the node cannot be identified.
    assert len(node.ops) == 1

    nodemeta = {
        "lineno": node.lineno,
        "col_offset": node.col_offset,
    }
    if PY311:
        return ast.Call(
            func=ast.Attribute(
                value=node.left,
                attr=CMP2MAGIC[type(node.ops[0])],
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[node.comparators[0]],
            keywords=[],
        )
    else:  # pragma: no cover
        return ast.Call(  # type: ignore
            func=ast.Attribute(
                value=node.left,
                attr=CMP2MAGIC[type(node.ops[0])],
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[node.comparators[0]],
            keywords=[],
            starargs=None,
            kwargs=None,
        )


@reconstruct_func_node.register(ast.BinOp)
def _(node: ast.BinOp) -> ast.Call:
    """Reconstruct the function node for `x + a`"""
    nodemeta = {
        "lineno": node.lineno,
        "col_offset": node.col_offset,
    }

    if PY311:
        return ast.Call(
            func=ast.Attribute(
                value=node.left,
                attr=OP2MAGIC[type(node.op)],
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[node.right],
            keywords=[],
        )
    else:  # pragma: no cover
        return ast.Call(  # type: ignore
            func=ast.Attribute(
                value=node.left,
                attr=OP2MAGIC[type(node.op)],
                ctx=ast.Load(),
                **nodemeta,
            ),
            args=[node.right],
            keywords=[],
            starargs=None,
            kwargs=None,
        )


def rich_exc_message(msg: str, node: ast.AST, context_lines: int = 4) -> str:
    """Attach the source code from the node to message to
    get a rich message for exceptions

    If package 'rich' is not install or 'node.__frame__' doesn't exist, fall
    to plain message (with basic information), otherwise show a better message
    with full information
    """
    frame = node.__frame__  # type: FrameType
    lineno = node.lineno - 1  # type: int
    col_offset = node.col_offset  # type: int
    filename = frame.f_code.co_filename  # type: str
    try:
        lines, startlineno = inspect.getsourcelines(frame)
    except OSError:  # pragma: no cover
        # could not get source code
        return f"{msg}\n"
    startlineno = 0 if startlineno == 0 else startlineno - 1
    line_range = (startlineno + 1, startlineno + len(lines) + 1)

    linenos = tuple(map(str, range(*line_range)))  # type: Tuple[str, ...]
    lineno_width = max(map(len, linenos))  # type: int
    hiline = lineno - startlineno  # type: int
    codes = []  # type: List[str]
    for i, lno in enumerate(linenos):
        if i < hiline - context_lines or i > hiline + context_lines:
            continue
        lno = lno.ljust(lineno_width)
        if i == hiline:
            codes.append(f"  > | {lno}  {lines[i]}")
            codes.append(f"    | {' ' * (lineno_width + col_offset + 2)}^\n")
        else:
            codes.append(f"    | {lno}  {lines[i]}")

    return (
        f"{msg}\n\n"
        f"  {filename}:{lineno + 1}:{col_offset + 1}\n"
        f"{''.join(codes)}\n"
    )



"""Provide core features for varname"""
from __future__ import annotations
import ast
import re
import warnings
from typing import Any, List, Union, Tuple, Type, Callable, overload

from executing import Source

from .utils import (
    bytecode_nameof,
    get_node,
    get_node_by_frame,
    lookfor_parent_assign,
    node_name,
    get_argument_sources,
    get_function_called_argname,
    rich_exc_message,
    reconstruct_func_node,
    ArgSourceType,
    VarnameRetrievingError,
    ImproperUseError,
    MultiTargetAssignmentWarning,
)
from .ignore import IgnoreList, IgnoreType


def varname(
    frame: int = 1,
    ignore: IgnoreType = None,
    multi_vars: bool = False,
    raise_exc: bool = True,
    strict: bool = True,
) -> Union[str, Tuple[Union[str, Tuple], ...]]:
    """Get the name of the variable(s) that assigned by function call or
    class instantiation.

    To debug and specify the right frame and ignore arguments, you can set
    debug on and see how the frames are ignored or selected:

    >>> from varname import config
    >>> config.debug = True

    Args:
        frame: `N`th frame used to retrieve the variable name. This means
            `N-1` intermediate frames will be skipped. Note that the frames
            match `ignore` will not be counted. See `ignore` for details.
        ignore: Frames to be ignored in order to reach the `N`th frame.
            These frames will not be counted to skip within that `N-1` frames.
            You can specify:
            - A module (or filename of a module). Any calls from it and its
                submodules will be ignored.
            - A function. If it looks like it might be a decorated function,
                a `MaybeDecoratedFunctionWarning` will be shown.
            - Tuple of a function and a number of additional frames that should
                be skipped just before reaching this function in the stack.
                This is typically used for functions that have been decorated
                with a 'classic' decorator that replaces the function with
                a wrapper. In that case each such decorator involved should
                be counted in the number that's the second element of the tuple.
            - Tuple of a module (or filename) and qualified name (qualname).
                You can use Unix shell-style wildcards to match the qualname.
                Otherwise the qualname must appear exactly once in the
                module/file.
            By default, all calls from `varname` package, python standard
            libraries and lambda functions are ignored.
        multi_vars: Whether allow multiple variables on left-hand side (LHS).
            If `True`, this function returns a tuple of the variable names,
            even there is only one variable on LHS.
            If `False`, and multiple variables on LHS, a
            `ImproperUseError` will be raised.
        raise_exc: Whether we should raise an exception if failed
            to retrieve the ast node.
            Note that set this to `False` will NOT supress the exception when
            the use of `varname` is improper (i.e. multiple variables on
            LHS with `multi_vars` is `False`). See `Raises/ImproperUseError`.
        strict: Whether to only return the variable name(s) if the result of
            the call is assigned to it/them directly. For example, `a = func()`
            rather than `a = [func()]`

    Returns:
        The variable name, or `None` when `raise_exc` is `False` and
            we failed to retrieve the ast node for the variable(s).
        A tuple or a hierarchy (tuple of tuples) of variable names
            when `multi_vars` is `True`.

    Raises:
        VarnameRetrievingError: When we are unable to retrieve the ast node
            for the variable(s) and `raise_exc` is set to `True`.

        ImproperUseError: When the use of `varname()` is improper, including:
            - When LHS is not an `ast.Name` or `ast.Attribute` node or not a
                list/tuple of them
            - When there are multiple variables on LHS but `multi_vars` is False
            - When `strict` is True, but the result is not assigned to
                variable(s) directly

            Note that `raise_exc=False` will NOT suppress this exception.

        MultiTargetAssignmentWarning: When there are multiple target
            in the assign node. (e.g: `a = b = func()`, in such a case,
            `a == 'b'`, may not be the case you want)
    """
    # Skip one more frame, as it is supposed to be called
    # inside another function
    refnode = get_node(frame + 1, ignore, raise_exc=raise_exc)
    if not refnode:
        if raise_exc:
            raise VarnameRetrievingError("Unable to retrieve the ast node.")
        return None

    node = lookfor_parent_assign(refnode, strict=strict)
    if not node:  # improper use
        if strict:
            msg = "Caller doesn't assign the result directly to variable(s)."
        else:
            msg = "Expression is not part of an assignment."

        raise ImproperUseError(rich_exc_message(msg, refnode))

    if isinstance(node, ast.Assign):
        # Need to actually check that there's just one
        # give warnings if: a = b = func()
        if len(node.targets) > 1:
            warnings.warn(
                "Multiple targets in assignment, variable name "
                "on the very right is used. ",
                MultiTargetAssignmentWarning,
            )
        target = node.targets[-1]
    else:
        target = node.target

    names = node_name(target)

    if not isinstance(names, tuple):
        names = (names,)

    if multi_vars:
        return names

    if len(names) > 1:
        raise ImproperUseError(
            rich_exc_message(
                "Expect a single variable on left-hand side, "
                f"got {len(names)}.",
                refnode,
            )
        )

    return names[0]


def will(frame: int = 1, raise_exc: bool = True) -> str:
    """Detect the attribute name right immediately after a function call.

    Examples:
        >>> class AwesomeClass:
        >>>     def __init__(self):
        >>>         self.will = None

        >>>     def permit(self):
        >>>         self.will = will()
        >>>         if self.will == 'do':
        >>>             # let self handle do
        >>>             return self
        >>>         raise AttributeError(
        >>>             'Should do something with AwesomeClass object'
        >>>         )

        >>>     def do(self):
        >>>         if self.will != 'do':
        >>>             raise AttributeError("You don't have permission to do")
        >>>         return 'I am doing!'

        >>> awesome = AwesomeClass()
        >>> # AttributeError: You don't have permission to do
        >>> awesome.do()
        >>> # AttributeError: Should do something with AwesomeClass object
        >>> awesome.permit()
        >>> awesome.permit().do() == 'I am doing!'

    Args:
        frame: At which frame this function is called.
        raise_exc: Raise exception we failed to detect the ast node
            This will NOT supress the `ImproperUseError`

    Returns:
        The attribute name right after the function call.
        `None` if ast node cannot be retrieved and `raise_exc` is `False`

    Raises:
        VarnameRetrievingError: When `raise_exc` is `True` and we failed to
            detect the attribute name (including not having one)

        ImproperUseError: When (the wraper of) this function is not called
            inside a method/property of a class instance.
            Note that this exception will not be suppressed by `raise_exc=False`
    """
    node = get_node(frame + 1, raise_exc=raise_exc)
    if not node:
        if raise_exc:
            raise VarnameRetrievingError("Unable to retrieve the frame.")
        return None

    # try to get node inst.attr from inst.attr()
    node = node.parent

    # see test_will_fail
    if not isinstance(node, ast.Attribute):
        if raise_exc:
            raise ImproperUseError(
                "Function `will` has to be called within "
                "a method/property of a class."
            )
        return None
    # ast.Attribute
    return node.attr


@overload
def nameof(
    var: Any,
    *,
    frame: int = 1,
    vars_only: bool = True,
) -> str:  # pragma: no cover
    ...


@overload
def nameof(
    var: Any,
    more_var: Any,
    /,  # introduced in python 3.8
    *more_vars: Any,
    frame: int = 1,
    vars_only: bool = True,
) -> Tuple[str, ...]:  # pragma: no cover
    ...


def nameof(
    var: Any,
    *more_vars: Any,
    frame: int = 1,
    vars_only: bool = True,
) -> Union[str, Tuple[str, ...]]:
    """Get the names of the variables passed in

    Examples:
        >>> a = 1
        >>> nameof(a) # 'a'

        >>> b = 2
        >>> nameof(a, b) # ('a', 'b')

        >>> x = lambda: None
        >>> x.y = 1
        >>> nameof(x.y, vars_only=False) # 'x.y'

    Note:
        This function works with the environments where source code is
        available, in other words, the callee's node can be retrieved by
        `executing`. In some cases, for example, running code from python
        shell/REPL or from `exec`/`eval`, we try to fetch the variable name
        from the bytecode. This requires only a single variable name is passed
        to this function and no keyword arguments, meaning that getting full
        names of attribute calls are not supported in such cases.

    Args:
        var: The variable to retrieve the name of
        *more_vars: Other variables to retrieve the names of
        frame: The this function is called from the wrapper of it. `frame=1`
            means no wrappers.
            Note that the calls from standard libraries are ignored.
            Also note that the wrapper has to have signature as this one.
        vars_only: Whether only allow variables/attributes as arguments or
            any expressions. If `False`, then the sources of the arguments
            will be returned.

    Returns:
        The names/sources of variables/expressions passed in.
            If a single argument is passed, return the name/source of it.
            If multiple variables are passed, return a tuple of their
            names/sources.
            If the argument is an attribute (e.g. `a.b`) and `vars_only` is
            `True`, only `"b"` will returned. Set `vars_only` to `False` to
            get `"a.b"`.

    Raises:
        VarnameRetrievingError: When the callee's node cannot be retrieved or
            trying to retrieve the full name of non attribute series calls.
    """
    # Frame is anyway used in get_node
    frameobj = IgnoreList.create(
        ignore_lambda=False,
        ignore_varname=False,
    ).get_frame(frame)

    node = get_node_by_frame(frameobj, raise_exc=True)
    if not node:
        # We can't retrieve the node by executing.
        # It can be due to running code from python/shell, exec/eval or
        # other environments where sourcecode cannot be reached
        # make sure we keep it simple (only single variable passed and no
        # full passed) to use bytecode_nameof
        #
        # We don't have to check keyword arguments here, as the instruction
        # will then be CALL_FUNCTION_KW.
        if not more_vars:
            return bytecode_nameof(frameobj.f_code, frameobj.f_lasti)

        # We are anyway raising exceptions, no worries about additional burden
        # of frame retrieval again
        source = frameobj.f_code.co_filename
        if source == "<stdin>":
            raise VarnameRetrievingError(
                "Are you trying to call nameof in REPL/python shell? "
                "In such a case, nameof can only be called with single "
                "argument and no keyword arguments."
            )
        if source == "<string>":
            raise VarnameRetrievingError(
                "Are you trying to call nameof from exec/eval? "
                "In such a case, nameof can only be called with single "
                "argument and no keyword arguments."
            )
        raise VarnameRetrievingError(
            "Source code unavailable, nameof can only retrieve the name of "
            "a single variable, and argument `full` should not be specified."
        )

    out = argname(
        "var",
        "*more_vars",
        func=nameof,
        frame=frame,
        vars_only=vars_only,
    )
    return out if more_vars else out[0]  # type: ignore


@overload
def argname(
    arg: str,
    *,
    func: Callable = None,
    dispatch: Type = None,
    frame: int = 1,
    ignore: IgnoreType = None,
    vars_only: bool = True,
) -> ArgSourceType:  # pragma: no cover
    ...


@overload
def argname(
    arg: str,
    more_arg: str,
    /,  # introduced in python 3.8
    *more_args: str,
    func: Callable = None,
    dispatch: Type = None,
    frame: int = 1,
    ignore: IgnoreType = None,
    vars_only: bool = True,
) -> Tuple[ArgSourceType, ...]:  # pragma: no cover
    ...


def argname(
    arg: str,
    *more_args: str,
    func: Callable = None,
    dispatch: Type = None,
    frame: int = 1,
    ignore: IgnoreType = None,
    vars_only: bool = True,
) -> Union[ArgSourceType, Tuple[ArgSourceType, ...]]:
    """Get the names/sources of arguments passed to a function.

    Instead of passing the argument variables themselves to this function
    (like `argname()` does), you should pass their names instead.

    Args:
        arg: and
        *more_args: The names of the arguments that you want to retrieve
            names/sources of.
            You can also use subscripts to get parts of the results.
            >>> def func(*args, **kwargs):
            >>>     return argname('args[0]', 'kwargs[x]') # no quote needed

            Star argument is also allowed:
            >>> def func(*args, x = 1):
            >>>     return argname('*args', 'x')
            >>> a = b = c = 1
            >>> func(a, b, x=c) # ('a', 'b', 'c')

            Note the difference:
            >>> def func(*args, x = 1):
            >>>     return argname('args', 'x')
            >>> a = b = c = 1
            >>> func(a, b, x=c) # (('a', 'b'), 'c')

        func: The target function. If not provided, the AST node of the
            function call will be used to fetch the function:
            - If a variable (ast.Name) used as function, the `node.id` will
                be used to get the function from `locals()` or `globals()`.
            - If variable (ast.Name), attributes (ast.Attribute),
                subscripts (ast.Subscript), and combinations of those and
                literals used as function, `pure_eval` will be used to evaluate
                the node
            - If `pure_eval` is not installed or failed to evaluate, `eval`
                will be used. A warning will be shown since unwanted side
                effects may happen in this case.
            You are very encouraged to always pass the function explicitly.
        dispatch: If a function is a single-dispatched function, you can
            specify a type for it to dispatch the real function. If this is
            specified, expect `func` to be the generic function if provided.
        frame: The frame where target function is called from this call.
            Calls from python standard libraries are ignored.
        ignore: The intermediate calls to be ignored. See `varname.ignore`
        vars_only: Require the arguments to be variables only.
            If False, `asttokens` is required to retrieve the source.

    Returns:
        The argument source when no more_args passed, otherwise a tuple of
        argument sources
        Note that when an argument is an `ast.Constant`, `repr(arg.value)`
        is returned, so `argname()` return `'a'` for `func("a")`

    Raises:
        VarnameRetrievingError: When the ast node where the function is called
            cannot be retrieved
        ImproperUseError: When frame or func is incorrectly specified.
    """
    ignore_list = IgnoreList.create(
        ignore,
        ignore_lambda=False,
        ignore_varname=False,
    )
    # where func(...) is called, skip the argname() call
    func_frame = ignore_list.get_frame(frame + 1)
    func_node = get_node_by_frame(func_frame)
    # Only do it when func_node are available
    if not func_node:
        # We can do something at bytecode level, when a single positional
        # argument passed to both functions (argname and the target function)
        # However, it's hard to ensure that there is only a single positional
        # arguments passed to the target function, at bytecode level.
        raise VarnameRetrievingError(
            "Cannot retrieve the node where the function is called."
        )

    func_node = reconstruct_func_node(func_node)

    if not func:
        func = get_function_called_argname(func_frame, func_node)

    if dispatch:
        func = func.dispatch(dispatch)

    # don't pass the target arguments so that we can cache the sources in
    # the same call. For example:
    # >>> def func(a, b):
    # >>>   a_name = argname(a)
    # >>>   b_name = argname(b)
    try:
        argument_sources = get_argument_sources(
            Source.for_frame(func_frame),
            func_node,
            func,
            vars_only=vars_only,
        )
    except Exception as err:
        raise ImproperUseError(
            "Have you specified the right `frame` or `func`?"
        ) from err

    out: List[ArgSourceType] = []
    farg_star = False
    for farg in (arg, *more_args):

        farg_name = farg
        farg_subscript = None  # type: str | int
        match = re.match(r"^([\w_]+)\[(.+)\]$", farg)
        if match:
            farg_name = match.group(1)
            farg_subscript = match.group(2)
            if farg_subscript.isdigit():
                farg_subscript = int(farg_subscript)
        else:
            match = re.match(r"^\*([\w_]+)$", farg)
            if match:
                farg_name = match.group(1)
                farg_star = True

        if farg_name not in argument_sources:
            raise ImproperUseError(
                f"{farg_name!r} is not a valid argument "
                f"of {func.__qualname__!r}."
            )

        source = argument_sources[farg_name]
        if isinstance(source, ast.AST):
            raise ImproperUseError(
                f"Argument {ast.dump(source)} is not a variable "
                "or an attribute."
            )

        if isinstance(farg_subscript, int) and not isinstance(source, tuple):
            raise ImproperUseError(
                f"`{farg_name}` is not a positional argument."
            )

        if isinstance(farg_subscript, str) and not isinstance(source, dict):
            raise ImproperUseError(
                f"`{farg_name}` is not a keyword argument."
            )

        if farg_subscript is not None:
            out.append(source[farg_subscript])  # type: ignore
        elif farg_star:
            out.extend(source)
        else:
            out.append(source)

    return (
        out[0]
        if not more_args and not farg_star
        else tuple(out)  # type: ignore
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['varname']

package_data = \
{'': ['*']}

install_requires = \
['executing>=2.1,<3.0']

extras_require = \
{':python_version < "3.10"': ['typing_extensions>=4.13,<5.0'],
 'all': ['asttokens==3.*', 'pure_eval==0.*']}

setup_kwargs = {
    'name': 'varname',
    'version': '0.15.1',
    'description': 'Dark magics about variable names in python.',
    'long_description': '![varname][7]\n\n[![Pypi][3]][4] [![Github][5]][6] [![PythonVers][8]][4] ![Building][10]\n[![Docs and API][9]][15] [![Codacy][12]][13] [![Codacy coverage][14]][13]\n![Downloads][17]\n\nDark magics about variable names in python\n\n[CHANGELOG][16] | [API][15] | [Playground][11] | :fire: [StackOverflow answer][20]\n\n## Installation\n\n```shell\npip install -U varname\n```\n\nNote if you use `python < 3.8`, install `varname < 0.11`\n\n## Features\n\n- Core features:\n\n  - Retrieving names of variables a function/class call is assigned to from inside it, using `varname`.\n  - Detecting next immediate attribute name, using `will`\n  - Fetching argument names/sources passed to a function using `argname`\n\n- Other helper APIs (built based on core features):\n\n  - A value wrapper to store the variable name that a value is assigned to, using `Wrapper`\n  - A decorator to register `__varname__` to functions/classes, using `register`\n  - A helper function to create dict without explicitly specifying the key-value pairs, using `jsobj`\n  - A `debug` function to print variables with their names and values\n  - `exec_code` to replace `exec` where source code is available at runtime\n\n## Credits\n\nThanks goes to these awesome people/projects:\n\n<table>\n  <tr>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/alexmojaki/executing">\n        <img src="https://ui-avatars.com/api/?color=3333ff&background=ffffff&bold=true&name=e&size=400" width="50px;" alt=""/>\n        <br /><sub><b>executing</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/alexmojaki">\n        <img src="https://avatars0.githubusercontent.com/u/3627481?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@alexmojaki</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/breuleux">\n        <img src="https://avatars.githubusercontent.com/u/599820?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@breuleux</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/ElCuboNegro">\n        <img src="https://avatars.githubusercontent.com/u/5524219?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@ElCuboNegro</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/thewchan">\n        <img src="https://avatars.githubusercontent.com/u/49702524?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@thewchan</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/LawsOfSympathy">\n        <img src="https://avatars.githubusercontent.com/u/96355982?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@LawsOfSympathy</b></sub>\n      </a>\n    </td>\n    <td align="center" style="min-width: 75px">\n      <a href="https://github.com/elliotgunton">\n        <img src="https://avatars.githubusercontent.com/u/17798778?s=400&v=4" width="50px;" alt=""/>\n        <br /><sub><b>@elliotgunton</b></sub>\n      </a>\n    </td>\n  </tr>\n</table>\n\nSpecial thanks to [@HanyuuLu][2] to give up the name `varname` in pypi for this project.\n\n## Usage\n\n### Retrieving the variable names using `varname(...)`\n\n- From inside a function\n\n    ```python\n    from varname import varname\n    def function():\n        return varname()\n\n    func = function()  # func == \'func\'\n    ```\n\n    When there are intermediate frames:\n\n    ```python\n    def wrapped():\n        return function()\n\n    def function():\n        # retrieve the variable name at the 2nd frame from this one\n        return varname(frame=2)\n\n    func = wrapped() # func == \'func\'\n    ```\n\n    Or use `ignore` to ignore the wrapped frame:\n\n    ```python\n    def wrapped():\n        return function()\n\n    def function():\n        return varname(ignore=wrapped)\n\n    func = wrapped() # func == \'func\'\n    ```\n\n    Calls from standard libraries are ignored by default:\n\n    ```python\n    import asyncio\n\n    async def function():\n        return varname()\n\n    func = asyncio.run(function()) # func == \'func\'\n    ```\n\n    Use `strict` to control whether the call should be assigned to\n    the variable directly:\n\n    ```python\n    def function(strict):\n        return varname(strict=strict)\n\n    func = function(True)     # OK, direct assignment, func == \'func\'\n\n    func = [function(True)]   # Not a direct assignment, raises ImproperUseError\n    func = [function(False)]  # OK, func == [\'func\']\n\n    func = function(False), function(False)   # OK, func = (\'func\', \'func\')\n    ```\n\n- Retrieving name of a class instance\n\n    ```python\n    class Foo:\n        def __init__(self):\n            self.id = varname()\n\n        def copy(self):\n            # also able to fetch inside a method call\n            copied = Foo() # copied.id == \'copied\'\n            copied.id = varname() # assign id to whatever variable name\n            return copied\n\n    foo = Foo()   # foo.id == \'foo\'\n\n    foo2 = foo.copy() # foo2.id == \'foo2\'\n    ```\n\n- Multiple variables on Left-hand side\n\n    ```python\n    # since v0.5.4\n    def func():\n        return varname(multi_vars=True)\n\n    a = func() # a == (\'a\',)\n    a, b = func() # (a, b) == (\'a\', \'b\')\n    [a, b] = func() # (a, b) == (\'a\', \'b\')\n\n    # hierarchy is also possible\n    a, (b, c) = func() # (a, b, c) == (\'a\', \'b\', \'c\')\n    ```\n\n- Some unusual use\n\n    ```python\n    def function(**kwargs):\n        return varname(strict=False)\n\n    func = func1 = function()  # func == func1 == \'func1\'\n    # if varname < 0.8: func == func1 == \'func\'\n    # a warning will be shown\n    # since you may not want func to be \'func1\'\n\n    x = function(y = function())  # x == \'x\'\n\n    # get part of the name\n    func_abc = function()[-3:]  # func_abc == \'abc\'\n\n    # function alias supported now\n    function2 = function\n    func = function2()  # func == \'func\'\n\n    a = lambda: 0\n    a.b = function() # a.b == \'a.b\'\n    ```\n\n### The decorator way to register `__varname__` to functions/classes\n\n- Registering `__varname__` to functions\n\n    ```python\n    from varname.helpers import register\n\n    @register\n    def function():\n        return __varname__\n\n    func = function() # func == \'func\'\n    ```\n\n    ```python\n    # arguments also allowed (frame, ignore and raise_exc)\n    @register(frame=2)\n    def function():\n        return __varname__\n\n    def wrapped():\n        return function()\n\n    func = wrapped() # func == \'func\'\n    ```\n\n- Registering `__varname__` as a class property\n\n    ```python\n    @register\n    class Foo:\n        ...\n\n    foo = Foo()\n    # foo.__varname__ == \'foo\'\n    ```\n\n### Getting variable names directly using `nameof`\n\n```python\nfrom varname import varname, nameof\n\na = 1\nnameof(a) # \'a\'\n\nb = 2\nnameof(a, b) # (\'a\', \'b\')\n\ndef func():\n    return varname() + \'_suffix\'\n\nf = func() # f == \'f_suffix\'\nnameof(f)  # \'f\'\n\n# get full names of (chained) attribute calls\nfunc.a = func\nnameof(func.a, vars_only=False) # \'func.a\'\n\nfunc.a.b = 1\nnameof(func.a.b, vars_only=False) # \'func.a.b\'\n```\n\n### Detecting next immediate attribute name\n\n```python\nfrom varname import will\nclass AwesomeClass:\n    def __init__(self):\n        self.will = None\n\n    def permit(self):\n        self.will = will(raise_exc=False)\n        if self.will == \'do\':\n            # let self handle do\n            return self\n        raise AttributeError(\'Should do something with AwesomeClass object\')\n\n    def do(self):\n        if self.will != \'do\':\n            raise AttributeError("You don\'t have permission to do")\n        return \'I am doing!\'\n\nawesome = AwesomeClass()\nawesome.do() # AttributeError: You don\'t have permission to do\nawesome.permit() # AttributeError: Should do something with AwesomeClass object\nawesome.permit().do() == \'I am doing!\'\n```\n\n### Fetching argument names/sources using `argname`\n\n```python\nfrom varname import argname\n\ndef func(a, b=1):\n    print(argname(\'a\'))\n\nx = y = z = 2\nfunc(x) # prints: x\n\n\ndef func2(a, b=1):\n    print(argname(\'a\', \'b\'))\nfunc2(y, b=x) # prints: (\'y\', \'x\')\n\n\n# allow expressions\ndef func3(a, b=1):\n    print(argname(\'a\', \'b\', vars_only=False))\nfunc3(x+y, y+x) # prints: (\'x+y\', \'y+x\')\n\n\n# positional and keyword arguments\ndef func4(*args, **kwargs):\n    print(argname(\'args[1]\', \'kwargs[c]\'))\nfunc4(y, x, c=z) # prints: (\'x\', \'z\')\n\n\n# As of 0.9.0 (see: https://pwwang.github.io/python-varname/CHANGELOG/#v090)\n# Can also fetch the source of the argument for\n# __getattr__/__getitem__/__setattr/__setitem__/__add__/__lt__, etc.\nclass Foo:\n    def __setattr__(self, name, value):\n        print(argname("name", "value", func=self.__setattr__))\n\nFoo().a = 1 # prints: ("\'a\'", \'1\')\n\n```\n\n### Value wrapper\n\n```python\nfrom varname.helpers import Wrapper\n\nfoo = Wrapper(True)\n# foo.name == \'foo\'\n# foo.value == True\nbar = Wrapper(False)\n# bar.name == \'bar\'\n# bar.value == False\n\ndef values_to_dict(*args):\n    return {val.name: val.value for val in args}\n\nmydict = values_to_dict(foo, bar)\n# {\'foo\': True, \'bar\': False}\n```\n\n### Creating dictionary using `jsobj`\n\n```python\nfrom varname.helpers import jsobj\n\na = 1\nb = 2\njsobj(a, b) # {\'a\': 1, \'b\': 2}\njsobj(a, b, c=3) # {\'a\': 1, \'b\': 2, \'c\': 3}\n```\n\n### Debugging with `debug`\n\n```python\nfrom varname.helpers import debug\n\na = \'value\'\nb = [\'val\']\ndebug(a)\n# "DEBUG: a=\'value\'\\n"\ndebug(b)\n# "DEBUG: b=[\'val\']\\n"\ndebug(a, b)\n# "DEBUG: a=\'value\'\\nDEBUG: b=[\'val\']\\n"\ndebug(a, b, merge=True)\n# "DEBUG: a=\'value\', b=[\'val\']\\n"\ndebug(a, repr=False, prefix=\'\')\n# \'a=value\\n\'\n# also debug an expression\ndebug(a+a)\n# "DEBUG: a+a=\'valuevalue\'\\n"\n# If you want to disable it:\ndebug(a+a, vars_only=True) # ImproperUseError\n```\n\n### Replacing `exec` with `exec_code`\n\n```python\nfrom varname import argname\nfrom varname.helpers import exec_code\n\nclass Obj:\n    def __init__(self):\n        self.argnames = []\n\n    def receive(self, arg):\n        self.argnames.append(argname(\'arg\', func=self.receive))\n\nobj = Obj()\n# exec(\'obj.receive(1)\')  # Error\nexec_code(\'obj.receive(1)\')\nexec_code(\'obj.receive(2)\')\nobj.argnames # [\'1\', \'2\']\n```\n\n## Reliability and limitations\n\n`varname` is all depending on `executing` package to look for the node.\nThe node `executing` detects is ensured to be the correct one (see [this][19]).\n\nIt partially works with environments where other AST magics apply, including [`exec`][24] function,\n[`macropy`][21], [`birdseye`][22], [`reticulate`][23] with `R`, etc. Neither\n`executing` nor `varname` is 100% working with those environments. Use\nit at your own risk.\n\nFor example:\n\n- This will not work:\n\n  ```python\n  from varname import argname\n\n  def getname(x):\n      print(argname("x"))\n\n  a = 1\n  exec("getname(a)")  # Cannot retrieve the node where the function is called.\n\n  ## instead\n  # from varname.helpers import exec_code\n  # exec_code("getname(a)")\n  ```\n\n[1]: https://github.com/pwwang/python-varname\n[2]: https://github.com/HanyuuLu\n[3]: https://img.shields.io/pypi/v/varname?style=flat-square\n[4]: https://pypi.org/project/varname/\n[5]: https://img.shields.io/github/tag/pwwang/python-varname?style=flat-square\n[6]: https://github.com/pwwang/python-varname\n[7]: logo.png\n[8]: https://img.shields.io/pypi/pyversions/varname?style=flat-square\n[9]: https://img.shields.io/github/actions/workflow/status/pwwang/python-varname/docs.yml?branch=master\n[10]: https://img.shields.io/github/actions/workflow/status/pwwang/python-varname/build.yml?branch=master\n[11]: https://mybinder.org/v2/gh/pwwang/python-varname/dev?filepath=playground%2Fplayground.ipynb\n[12]: https://img.shields.io/codacy/grade/6fdb19c845f74c5c92056e88d44154f7?style=flat-square\n[13]: https://app.codacy.com/gh/pwwang/python-varname/dashboard\n[14]: https://img.shields.io/codacy/coverage/6fdb19c845f74c5c92056e88d44154f7?style=flat-square\n[15]: https://pwwang.github.io/python-varname/api/varname\n[16]: https://pwwang.github.io/python-varname/CHANGELOG/\n[17]: https://img.shields.io/pypi/dm/varname?style=flat-square\n[19]: https://github.com/alexmojaki/executing#is-it-reliable\n[20]: https://stackoverflow.com/a/59364138/5088165\n[21]: https://github.com/lihaoyi/macropy\n[22]: https://github.com/alexmojaki/birdseye\n[23]: https://rstudio.github.io/reticulate/\n[24]: https://docs.python.org/3/library/functions.html#exec\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pwwang/python-varname',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
